"""Test script for video class
"""
import sys
import os.path
import pytest
import cv2 as cv2
import numpy as np
import scipy
import skimage.transform
import multiprocessing

LIBDIR = os.path.dirname(os.path.abspath(__file__))
modules_path = os.path.join(os.path.split(LIBDIR)[0], "Katna")
test_path = os.path.join(os.path.split(LIBDIR)[0], "tests")
sys.path.append(modules_path)
sys.path.append(test_path)


@pytest.fixture(scope="module")
def video_object():
    """fixture for video object
    
    Returns:
        Video -- an instantiated video object
    """
    from video import Video

    if __name__ == "__main__":
        multiprocessing.set_start_method("spawn")

    return Video()


@pytest.fixture
def image_similarity_object():
    """fixture for video object
    
    Returns:
        Image_Similarity -- an instantiated Image_Similarity object
    """
    from image_similarity import ImageSimilarity

    return ImageSimilarity()


def test_validate_video_exception():
    """
        Test case for no file exception.
    """
    # Use following line to print anything from test function inside
    # with capsys.disabled():
    video_file_path = os.path.join(".", "data", "pos_video1.mp4")

    from decorators import FileDecorators

    @FileDecorators.validate_file_path
    def dummy(file_path):  # Dummy function because we only want to test the decorator
        return True

    with pytest.raises(FileNotFoundError):
        assert FileDecorators.validate_file_path(dummy(file_path=video_file_path))


def test_extracted_frame_numbers(video_object):
    """
        Test case for extracted frame numbers.
        Must return 10 images
    """
    # Use following line to print anything from test function inside
    # with capsys.disabled():
    video_file_path = os.path.join("tests", "data", "pos_video.mp4")
    expected_number_of_images = 10

    imgs = video_object.extract_video_keyframes(
        expected_number_of_images, video_file_path
    )

    assert len(imgs) == expected_number_of_images


def test_extracted_frame_quality(video_object, image_similarity_object):
    """
        Test case for extracted frame numbers.
        Must return similar 12 images as stored in test data 
    """
    # Use following line to print anything from test function inside
    # with capsys.disabled():
    video_file_path = os.path.join("tests", "data", "pos_video.mp4")
    expected_number_of_images = 12

    imgs = video_object.extract_video_keyframes(
        expected_number_of_images, video_file_path
    )

    print("Testing if extracted images are same as stored images inside data folder")

    for count in range(expected_number_of_images):
        test_img_path = os.path.join(
            "tests", "data", "extracted_img_" + str(count) + ".jpeg"
        )
        test_img = cv2.imread(test_img_path)

        image_found = False
        for img in imgs:
            cv2.imwrite("temp.jpeg", img)
            similairty = image_similarity_object.pixel_sim(test_img_path, "temp.jpeg")
            if abs(similairty - 0.0) < 0.01:
                image_found = True
                break

        os.remove("temp.jpeg")
        assert image_found == True


def test_video_splitting(video_object):
    """Test case for splitting logic for videos.. Used tide ad video because of edge case
    """
    large_video_file_path = os.path.join("tests", "data", "Tidead.mp4")
    small_video_file_path = os.path.join("tests", "data", "pos_video.mp4")
    n_clips_large = video_object._split(large_video_file_path)
    n_clips_small = video_object._split(small_video_file_path)
    video_object._remove_clips(n_clips_large)
    video_object._remove_clips(n_clips_small)
    # print(len(n_clips))
    if multiprocessing.cpu_count() > 1:
        assert len(n_clips_large) > 1
    assert len(n_clips_small) == 1


@pytest.mark.skip(reason="no way of currently testing this")
def test_extracted_frame_as_png(video_object):
    """
        Test case for png extraction.
        All returned files must be png.
    """
    # Use following line to print anything from test function inside
    # with capsys.disabled():
    video_file_path = os.path.join(".", "data", "pos_video.mp4")
    expected_number_of_images = 10
    file_type = "png"

    imgs = video_object.extract_video_keyframes(
        expected_number_of_images, video_file_path, file_type
    )

    # All the images must have png extension

    assert False


@pytest.mark.skip(reason="no way of currently testing this")
def test_extracted_frame_as_jpg(video_object):
    """
        Test case for jpg extraction.
        All returned files must be jpg.
    """
    # Use following line to print anything from test function inside
    # with capsys.disabled():
    video_file_path = os.path.join(".", "data", "pos_video.mp4")
    expected_number_of_images = 10

    imgs = video_object.extract_video_keyframes(
        expected_number_of_images, video_file_path
    )

    # All the images must have jpg extension

    assert False


@pytest.mark.skip(reason="no way of currently testing this")
def test_extract_frames_as_numpy(video_object):
    """
        Test case for frame extraction as image.
        Must return 10 images
    """
    # Use following line to print anything from test function inside
    # with capsys.disabled():
    video_file_path = os.path.join(".", "data", "pos_video.mp4")
    expected_number_of_images = 10
    imgs = video_object.extract_frames_as_numpy(
        expected_number_of_images, video_file_path
    )

    # Check that returned images are numpy array

    assert False


def test_save_frame_to_disk_exception(video_object, tmpdir):
    """Test case for exception test while saving extracted image
    
    :param video_object: Video class object
    :type video_object: Video
    """

    img = "Str"  # None numpy array
    folder = tmpdir.mkdir("katna_test_dir")
    with pytest.raises(TypeError):
        assert video_object.save_frame_to_disk(
            img, file_path=folder, file_name="test", file_ext=".jpeg"
        )


def test_save_frame_to_disk(video_object, tmpdir):
    """Test case for saving image extracted frame on disk
    
    :param video_object: Video class object
    :type video_object: Video
    """

    # Save the file to a temporary location
    test_file_name = "23018877.jpg"
    test_file_location = os.path.join("tests", "data", test_file_name)
    img = cv2.imread(test_file_location, cv2.IMREAD_UNCHANGED)
    # tempdir is pytest provided fixture for temporary folders
    copy_location = tmpdir.mkdir("katna_test_dir")

    video_object.save_frame_to_disk(
        img, file_path=copy_location, file_name="test", file_ext=".jpeg"
    )

    assert os.path.isfile(os.path.join(copy_location, "test.jpeg")) == True



"""
.. module:: Katna.filterlist
    :platform: OS X
    :synopsis: This module creates list of filters for image module
"""

import os
import cv2
import pathlib
import numpy as np


class FilterList:
    """Adapter class for filters
    """
    # TODO #2
    def __init__(self):
        self.filters = []
        # Getting the absolute path for the directory in which all the filter modules are kept
        filters_modules_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "image_filters")
        filters_modules_list = os.listdir(filters_modules_path)

        # Iterating over each filter module, importing it and adding it to filters list
        for each_filter in filters_modules_list:

            # Checking if the file is a python file and in format: <filter name>_detector.py
            if each_filter.endswith(".py") and len(each_filter.split("_")) == 2:
                splitted_name = each_filter.split(".")
                
                # Creating the import statement as string and converting it into import command command
                exec(
                    str(
                        "from Katna.image_filters."
                        + splitted_name[0]
                        + " import "
                        + "".join([_.title() for _ in splitted_name[0].split("_")])
                    )
                )

                # Creating a dummy object for the imported filter module class and adding it to the filters list
                dummy_obj = eval(
                    "".join([_.title() for _ in splitted_name[0].split("_")]) + "()"
                )
                self.filters.append(dummy_obj)

    def get_filters(self):
        """ Function to get list of all the builtin filters

            :return: Returns list of filters
            :rtype: python list of filter objects
        """
        return self.filters



import sys
import os.path
import pytest
import cv2
import numpy as np
import inspect

LIBDIR = os.path.dirname(os.path.abspath(__file__))
modules_path = os.path.join(os.path.split(LIBDIR)[0], "Katna")
test_path = os.path.join(os.path.split(LIBDIR)[0], "tests")
sys.path.append(modules_path)
sys.path.append(test_path)


def process_images(path, temp_dir, feature, which):

    test_img_ext = ["jpg", "jpeg"]
    i = 0
    method_list = inspect.getmembers(eval("which"))
    for f in os.listdir(path):
        if f.endswith(test_img_ext[0]) or f.endswith(test_img_ext[1]):
            img = cv2.imread(os.path.join(path, f))
            smap = eval("feature." + method_list[-2][0] + "(img)")
            cv2.imwrite(os.path.join(temp_dir, f), smap)
            i += 1
    return i


@pytest.fixture(scope="module")
def image_similarity_object():
    """fixture for image similarity object
    
    Returns:
        Image_Similarity -- an instantiated Image_Similarity object
    """
    from image_similarity import ImageSimilarity

    return ImageSimilarity()


@pytest.fixture(scope="module")
def image_object():
    """fixture for image object
    
    Returns:
        Image -- an instantiated Image object
    """
    from Katna.image import Image

    return Image()


@pytest.fixture(scope="module")
def text_detector_object():
    """fixture for text detection object
    
    Returns:
        TextDetector -- an instantiated Text Detector object
    """
    from Katna.image_filters.text_detector import TextDetector

    return TextDetector()


def test_saliency(tmpdir_factory):
    from Katna.image_features.saliency_feature import SaliencyFeature

    sd = SaliencyFeature()
    path = os.path.join("tests", "data")
    save_path = "saliency_result_images"
    temp_dir = tmpdir_factory.mktemp(save_path)
    no_img_processed = process_images(path, temp_dir, sd, SaliencyFeature)
    assert len(temp_dir.listdir()) == no_img_processed


def test_edge(tmpdir_factory):
    from Katna.image_features.edge_feature import EdgeFeature

    ed = EdgeFeature()
    path = os.path.join("tests", "data")
    save_path = "edge_result_images"
    temp_dir = tmpdir_factory.mktemp(save_path)
    no_img_processed = process_images(path, temp_dir, ed, EdgeFeature)
    assert len(temp_dir.listdir()) == no_img_processed


def test_face(tmpdir_factory):
    from Katna.image_features.face_feature import FaceFeature

    sf = FaceFeature()
    path = os.path.join("tests", "data")
    save_path = "face_result_images"
    temp_dir = tmpdir_factory.mktemp(save_path)
    no_img_processed = process_images(path, temp_dir, sf, FaceFeature)
    assert len(temp_dir.listdir()) == no_img_processed


def test_features_list():
    from Katna.feature_list import FeatureList

    fl = FeatureList()
    assert len(fl.get_features()) == 3


def test_filters_list():
    from Katna.filter_list import FilterList

    fl = FilterList()
    assert len(fl.get_filters()) == 1


def test_text_detector(tmpdir_factory, text_detector_object):

    from Katna.crop_rect import CropRect

    crop1 = CropRect(0, 0, 100, 100)
    crop2 = CropRect(0, 0, 200, 200)
    crop3 = CropRect(0, 0, 1000, 600)
    path = os.path.join("tests", "data")
    save_path = "text_result_image"
    temp_dir = tmpdir_factory.mktemp(save_path)
    text_detector_object.set_image(cv2.imread(os.path.join(path, "image_for_text.png")))
    assert text_detector_object.get_filter_result(crop1) == False
    assert text_detector_object.get_filter_result(crop2) == False
    assert text_detector_object.get_filter_result(crop3) == True


def test_image(image_object):
    import os.path
    import cv2

    # number of images to be returned
    no_of_crops_to_returned = 3

    # crop dimensions
    crop_width = 800
    crop_height = 600

    crop_width1 = 200
    crop_height1 = 200
    # Filters
    filters = ["text"]
    # Image file path
    path = os.path.join("tests", "data")
    image_file_path = os.path.join(path, "test_image_for_text.jpeg")
    print(f"image_file_path = {image_file_path}")

    img = cv2.imread(image_file_path)
    crop_list = image_object.crop_image_from_cvimage(
        input_image=img,
        crop_width=crop_width,
        crop_height=crop_height,
        num_of_crops=no_of_crops_to_returned,
        filters=filters,
    )
    crop_list1 = image_object.crop_image_from_cvimage(
        input_image=img,
        crop_width=crop_width1,
        crop_height=crop_height1,
        num_of_crops=no_of_crops_to_returned,
        filters=filters,
    )
    print(len(crop_list1), len(crop_list))
    assert len(crop_list) == no_of_crops_to_returned
    assert len(crop_list1) == 0



def test_image_resize(image_object):
    import os.path
    import cv2


    # crop dimensions
    target_width = 800
    target_height = 600

    target_width1 = 200
    target_height1 = 200

    path = os.path.join("tests", "data")
    image_file_path = os.path.join(path, "test_image_for_text.jpeg")
    print(f"image_file_path = {image_file_path}")

    resized_image = image_object.resize_image(
        file_path=image_file_path,
        target_width=target_width,
        target_height=target_height,
    )
    resized_image1 = image_object.resize_image(
        file_path=image_file_path,
        target_width=target_width1,
        target_height=target_height1,
    )
    assert resized_image is not None
    assert resized_image1 is not None
    resized_image_height, resized_image_width, _ = resized_image.shape
    resized_image1_height, resized_image1_width, _ = resized_image1.shape
    assert resized_image_height == target_height
    assert resized_image_width == target_width
    assert resized_image1_height == target_height1
    assert resized_image1_width == target_width1

def test_crop_quality(tmpdir_factory, image_object, image_similarity_object):
    import os.path
    import cv2

    # number of images to be returned
    no_of_crops_to_returned = 3

    # crop dimentions
    crop_width = 500
    crop_height = 500

    # Image file path
    path = os.path.join("tests", "data")
    save_path = "result_image_crop"
    temp_dir = tmpdir_factory.mktemp(save_path)
    image_file_path = os.path.join(path, "23018877.jpg")
    print(f"image_file_path = {image_file_path}")

    img = cv2.imread(image_file_path)
    crop_list = image_object.crop_image(
        file_path=image_file_path,
        crop_width=crop_width,
        crop_height=crop_height,
        num_of_crops=no_of_crops_to_returned,
    )
    print(crop_list)
    image_object.save_crop_to_disk(crop_list[0], img, temp_dir, "crop_img", ".jpg")
    similairty = image_similarity_object.pixel_sim(
        os.path.join(path, "23018877_crop.jpg"), os.path.join(temp_dir, "crop_img.jpg")
    )
    print("Similarity: ", similairty)
    assert round(similairty, 1) <= 0.3


def test_validate_image_exception():
    """
        Test case for no file exception.
    """
    # Use following line to print anything from test function inside
    # with capsys.disabled():
    image_file_path = os.path.join(".", "data", "image_1.jpg")

    from decorators import FileDecorators

    @FileDecorators.validate_file_path
    def dummy(file_path):  # Dummy function because we only want to test the decorator
        return True

    with pytest.raises(FileNotFoundError):
        assert FileDecorators.validate_file_path(dummy(file_path=image_file_path))


def test_aspect_ratio(tmpdir_factory, image_object):
    import os.path
    import cv2

    # number of images to be returned
    no_of_crops_to_returned = 3

    # crop aspect ratio
    crop_aspect_ratio = "4:3"
    aspect_ratio = list(map(int, crop_aspect_ratio.split(":")))
    # Filters
    filters = []
    # Image file path
    path = os.path.join("tests", "data")
    save_path = "result_image_crop"
    temp_dir = tmpdir_factory.mktemp(save_path)
    image_file_path = os.path.join(path, "bird_img_for_crop.jpg")
    print(f"image_file_path = {image_file_path}")

    img = cv2.imread(image_file_path)
    crop_list = image_object.crop_image_with_aspect(
        file_path=image_file_path,
        crop_aspect_ratio=crop_aspect_ratio,
        num_of_crops=no_of_crops_to_returned,
        filters=filters,
    )
    ratio = str(aspect_ratio[0] / aspect_ratio[1])
    for crop in crop_list:
        cv2.imwrite(
            os.path.join(temp_dir, "cropped.jpeg"),
            img[crop.y : crop.y + crop.h, crop.x : crop.x + crop.w],
        )
        crop_img = cv2.imread(os.path.join(temp_dir, "cropped.jpeg"))
        value = str(crop_img.shape[1] / crop_img.shape[0])
        print(value[:4], ratio[:4])
        if value[:3] != ratio[:3]:
            assert False
    assert True



"""
.. module:: Katna.config
    :platform: Platfrom Independent
    :synopsis: This module defines some helpful configuration variables
"""


# # Configuration parameters for Image class
class Image:
    # default value by which image size to be reduces for processing
    down_sample_factor = 8
    # Debug flag
    DEBUG = False
    # Crop_height_reduction_factor_in_each_iterationnot found crop height
    # will be reduced by this multiplier/factor and search for candidate crops
    # is resumed.
    # Decreasing the height and width for crops while checking it don't get small by 1/(min_image_to_crop_factor) of image height/width
    min_image_to_crop_factor = 4
    crop_height_reduction_factor_in_each_iteration = 0.05


# # Configurations for Scoring crops for crop extractor
class CropScorer:
    detail_weight = 0.2  # default weight value for detail parameter
    edge_radius = 0.4  # default edge radius
    edge_weight = -20  # default edge weight
    outside_importance = (
        -0.5
    )  # default value to set if the pixel is outside crop rectangle
    rule_of_thirds = True  # boolean to set rule of third condition check
    saliency_bias = 0.2  # bias color value for saliency(+- error value)
    saliency_weight = 1.3  # default edge radius
    face_bias = 0.01  # bias color value for face(+- error value)
    face_weight = 3.4  # default weight value for face parameter
    rects_weight = 1  # default weight value for crop rectangles


# # Configurations for Text detection class
class TextDetector:
    # Min Confidence Threshold for Text detection model
    min_confidence = 0.9
    # Threshold for merging text detection boxes
    merge_threshold = 1
    # Name of Model files to be used for text detection
    frozen_weights = "frozen_east_text_detection.pb"
    # Location where model file will be downloaded
    cache_subdir = "models"
    # Layers Name for text detection
    layerNames = ["feature_fusion/Conv_7/Sigmoid", "feature_fusion/concat_3"]
    # Download Link for Text detection model
    model_download_link = "https://github.com/oyyd/frozen_east_text_detection.pb/raw/master/frozen_east_text_detection.pb"


# # Configurations for Edge Feature class
class EdgeFeature:
    # min edge threshold value
    min_val_threshold = 100
    # Max edge threshold value
    max_val_threshold = 200
    # aperture_size/size of Sobel kernel for canny edge detector
    ksize = 3


# # Configurations for Face detection Feature class
class FaceFeature:
    # Model file name to be used for face detection
    model_file = "res10_300x300_ssd_iter_140000_fp16.caffemodel"
    # Model definition file name to be used for face detetion
    prototxt_file = "deploy.prototxt"
    # Location where model file will be downloaded
    cache_subdir = "models"
    # Min Confidence Threshold for face detection model
    confidence = 0.5
    # Download Link for face detection model defintion file
    prototxt_download_link = "https://raw.githubusercontent.com/opencv/opencv/master/samples/dnn/face_detector/deploy.prototxt"
    # Download Link for face detection model
    modelfile_download_link = "https://raw.githubusercontent.com/opencv/opencv_3rdparty/dnn_samples_face_detector_20180205_fp16/res10_300x300_ssd_iter_140000_fp16.caffemodel"


# # Configuration parameters for Video class
class Video:
    # Debug flag
    DEBUG = False
    min_video_duration = 5.0

    # https://trac.ffmpeg.org/wiki/Encode/H.264

    # Keep this between 20 to 30 value
    video_compression_crf_parameter = 23
    video_compression_codec = "libx264"  # Currently "libx264 and  is supported"
    compression_output_file_extension = "mp4"


class ImageSelector:
    # Setting for optimum Brightness values
    min_brightness_value = 10.0
    max_brightness_value = 90.0
    brightness_step = 2.0
    # Setting for optimum Contrast/Entropy values
    min_entropy_value = 1.0
    max_entropy_value = 10.0
    entropy_step = 0.5


class FrameExtractor:
    # Setting local maxima criteria
    USE_LOCAL_MAXIMA = True
    # Lenght of sliding window taking difference
    len_window = 20
    # Chunk size of Images to be processed at a time in memory
    max_frames_in_chunk = 2500
    # Type of smoothening window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman' flat window will produce a moving average smoothing.
    window_type = "hanning"



"""
.. module:: Katna.frame_extractor
    :platform: OS X
    :synopsis: This module has functions related to key frame extraction 
"""

import cv2
import operator
import numpy as np
from scipy.signal import argrelextrema

import tempfile
import Katna.config as config

# Class to hold information about each frame
class Frame:
    """Class for storing frame ref
    """

    def __init__(self, frame, sum_abs_diff):
        self.frame = frame
        self.sum_abs_diff = sum_abs_diff


class FrameExtractor(object):
    """Class for extraction of key frames from video : based on sum of absolute differences in LUV colorspace from given video 
    """

    def __init__(self):

        # Setting local maxima criteria
        self.USE_LOCAL_MAXIMA = config.FrameExtractor.USE_LOCAL_MAXIMA
        # Lenght of sliding window taking difference
        self.len_window = config.FrameExtractor.len_window
        # Chunk size of Images to be processed at a time in memory
        self.max_frames_in_chunk = config.FrameExtractor.max_frames_in_chunk

    def __calculate_frame_difference(self, frame, curr_frame, prev_frame):
        """Function to calculate the difference between current frame and previous frame

        :param frame: frame from the video
        :type frame: numpy array
        :param curr_frame: current frame from the video in LUV format
        :type curr_frame: numpy array
        :param prev_frame: previous frame from the video in LUV format
        :type prev_frame: numpy array
        :return: difference count and frame if None is empty or undefined else None
        :rtype: tuple
        """

        if curr_frame is not None and prev_frame is not None:
            # Calculating difference between current and previous frame
            diff = cv2.absdiff(curr_frame, prev_frame)
            count = np.sum(diff)
            frame = Frame(frame, count)

            return count, frame
        return None

    def __process_frame(self, frame, prev_frame, frame_diffs, frames):
        """Function to calculate the difference between current frame and previous frame

        :param frame: frame from the video
        :type frame: numpy array
        :param prev_frame: previous frame from the video in LUV format
        :type prev_frame: numpy array
        :param frame_diffs: list of frame differences
        :type frame_diffs: list of int
        :param frames: list of frames
        :type frames: list of numpy array
        :return: previous frame and current frame
        :rtype: tuple
        """

        luv = cv2.cvtColor(frame, cv2.COLOR_BGR2LUV)
        curr_frame = luv
        # Calculating the frame difference for previous and current frame
        frame_diff = self.__calculate_frame_difference(frame, curr_frame, prev_frame)
        
        if frame_diff is not None:
            count, frame = frame_diff
            frame_diffs.append(count)
            frames.append(frame)
        prev_frame = curr_frame

        return prev_frame, curr_frame

    def __extract_all_frames_from_video__(self, videopath):
        """Generator function for extracting frames from a input video which are sufficiently different from each other, 
        and return result back as list of opencv images in memory

        :param videopath: inputvideo path
        :type videopath: `str`
        :return: Generator with extracted frames in max_process_frames chunks and difference between frames
        :rtype: generator object with content of type [numpy.ndarray, numpy.ndarray] 
        """
        cap = cv2.VideoCapture(str(videopath))

        ret, frame = cap.read()
        i = 1
        chunk_no = 0
        while ret:
            curr_frame = None
            prev_frame = None

            frame_diffs = []
            frames = []
            for _ in range(0, self.max_frames_in_chunk):
                if ret:
                    # Calling process frame function to calculate the frame difference and adding the difference 
                    # in **frame_diffs** list and frame to **frames** list
                    prev_frame, curr_frame = self.__process_frame(frame, prev_frame, frame_diffs, frames)
                    i = i + 1
                    ret, frame = cap.read()
                    # print(frame_count)
                else:
                    cap.release()
                    break
            chunk_no = chunk_no + 1
            yield frames, frame_diffs
        cap.release()

    def __get_frames_in_local_maxima__(self, frames, frame_diffs):
        """ Internal function for getting local maxima of key frames 
        This functions Returns one single image with strongest change from its vicinity of frames 
        ( vicinity defined using window length ) 

        :param object: base class inheritance
        :type object: class:`Object`
        :param frames: list of frames to do local maxima on
        :type frames: `list of images`
        :param frame_diffs: list of frame difference values 
        :type frame_diffs: `list of images`

        """
        extracted_key_frames = []
        diff_array = np.array(frame_diffs)
        # Normalizing the frame differences based on windows parameters
        sm_diff_array = self.__smooth__(diff_array, self.len_window)

        # Get the indexes of those frames which have maximum differences
        frame_indexes = np.asarray(argrelextrema(sm_diff_array, np.greater))[0]

        for frame_index in frame_indexes:
            extracted_key_frames.append(frames[frame_index - 1].frame)
        return extracted_key_frames

    def __smooth__(self, x, window_len, window=config.FrameExtractor.window_type):
        """smooth the data using a window with requested size.
        This method is based on the convolution of a scaled window with the signal.
        The signal is prepared by introducing reflected copies of the signal
        (with the window size) in both ends so that transient parts are minimized
        in the begining and end part of the output signal.
        example:
        import numpy as np
        t = np.linspace(-2,2,0.1)
        x = np.sin(t)+np.random.randn(len(t))*0.1
        y = smooth(x)
        see also:
        numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
        scipy.signal.lfilter
        
        :param x: the frame difference list
        :type x: numpy.ndarray
        :param window_len: the dimension of the smoothing window
        :type window_len: slidding window length
        :param window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman' flat window will produce a moving average smoothing.
        :type window: str
        :return: the smoothed signal
        :rtype: ndarray
        """
        # This function takes 
        if x.ndim != 1:
            raise (ValueError, "smooth only accepts 1 dimension arrays.")

        if x.size < window_len:
            raise (ValueError, "Input vector needs to be bigger than window size.")

        if window_len < 3:
            return x

        if not window in ["flat", "hanning", "hamming", "bartlett", "blackman"]:
            raise (
                ValueError,
                "Smoothing Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'",
            )

        # Doing row-wise merging of frame differences wrt window length. frame difference
        # by factor of two and subtracting the frame differences from index == window length in reverse direction
        s = np.r_[2 * x[0] - x[window_len:1:-1], x, 2 * x[-1] - x[-1:-window_len:-1]]

        if window == "flat":  # moving average
            w = np.ones(window_len, "d")
        else:
            w = getattr(np, window)(window_len)
        y = np.convolve(w / w.sum(), s, mode="same")
        return y[window_len - 1 : -window_len + 1]

    def extract_candidate_frames(self, videopath):
        """ Pubic function for this module , Given and input video path
        This functions Returns one list of all candidate key-frames  

        :param object: base class inheritance
        :type object: class:`Object`
        :param videopath: inputvideo path
        :type videopath: `str`
        :return: opencv.Image.Image objects
        :rtype: list
        """

        extracted_candidate_key_frames = []

        # Get all frames from video in chunks using python Generators
        frame_extractor_from_video_generator = self.__extract_all_frames_from_video__(
            videopath
        )

        # Loop over every frame in the frame extractor generator object and calculate the
        # local maxima of frames 
        for frames, frame_diffs in frame_extractor_from_video_generator:
            extracted_candidate_key_frames_chunk = []
            if self.USE_LOCAL_MAXIMA:

                # Getting the frame with maximum frame difference
                extracted_candidate_key_frames_chunk = self.__get_frames_in_local_maxima__(
                    frames, frame_diffs
                )
                extracted_candidate_key_frames.extend(
                    extracted_candidate_key_frames_chunk
                )

        return extracted_candidate_key_frames



"""
.. module:: Katna.image_features.saliency_feature
    :platform: Cross Platform
    :synopsis: This module is for for getting saliency feature Map from input image
"""

import os
import cv2
import numpy as np
from Katna.image_features.feature import Feature


class SaliencyFeature(Feature):
    """Class for calculating saliency feature map from Input image
    """

    def __init__(self, weight=0.0):
        # Initialize weight for feature in base class
        super().__init__(weight)
        self.saliency = cv2.saliency.StaticSaliencySpectralResidual_create()

    def get_feature_map(self, image):
        """Public function for getting Feature map image from Image saliency detection

        :param image: input image
        :type image: `numpy array`
        :return: single channel opencv numpy image with feature map from saliency detection
        :rtype: numpy array
        """
        _, saliency_map = self.saliency.computeSaliency(image)
        saliency_map = (saliency_map * 255).astype("uint8")

        return saliency_map



"""
.. module:: Katna.decorators
    :platform: OS X
    :synopsis: This module has decorators for video and image modules
"""
import os.path
import os
import sys
import errno
import functools
import inspect
import Katna.config as config


class VideoDecorators(object):
    """File validation decorator

    Arguments:
        object {[type]} -- [description]

    Raises:
        FileNotFoundError: [Video File is missing]

    Returns:
        [boolean] -- [if the file exists and is valid]
    """

    @classmethod
    def validate_video(cls, decorated):
        """Validate if the input video is a valid file
        """

        @functools.wraps(decorated)
        def wrapper(*args, **kwargs):
            """ wrapper for decorated function

            Arguments:
                cls {VideoDecorators} -- [Video decorators class]

            Raises:
                FileNotFoundError: [If the file is missing]

            Returns:
                [function] -- [Decorated function]
            """
            func_args = inspect.getcallargs(decorated, *args, **kwargs)

            key = "file_path"

            if key not in func_args:
                raise Exception("File_path parameter is missing")
            else:
                f_path = func_args.get(key)

            if bool(f_path is None or os.path.isfile(f_path) is False):
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), f_path)
            else:
                return decorated(*args, **kwargs)

        return wrapper


class FileDecorators(object):
    """File validation decorator

    :raises FileNotFoundError: File or path is incorrect
    """

    @classmethod
    def validate_dir_path(cls, decorated):
        """Validate if the input path is a valid dir or location

        :param decorated: decorated function
        :type decorated: function, required
        :return: function if the path is valid
        :rtype: function object
        """

        @functools.wraps(decorated)
        def wrapper(*args, **kwargs):
            """ wrapper for decorated function. args and kwargs are standard
            function parameters.
            """
            func_args = inspect.getcallargs(decorated, *args, **kwargs)

            key = "dir_path"

            if key not in func_args:
                raise Exception("dir_path parameter is missing")
            else:
                f_path = func_args.get(key)

            if bool(f_path is None or os.path.exists(f_path) is False):
                raise NotADirectoryError(errno.ENOENT, os.strerror(errno.ENOENT), f_path)
            else:
                return decorated(*args, **kwargs)

        return wrapper

    @classmethod
    def validate_file_path(cls, decorated):
        """Validate if the input path is a valid file or location

        :param decorated: decorated function
        :type decorated: function, required
        :return: function if the path is valid
        :rtype: function object
        """

        @functools.wraps(decorated)
        def wrapper(*args, **kwargs):
            """ wrapper for decorated function. args and kwargs are standard
            function parameters.
            """
            func_args = inspect.getcallargs(decorated, *args, **kwargs)

            key = "file_path"

            if key not in func_args:
                raise Exception("File_path parameter is missing")
            else:
                f_path = func_args.get(key)

            if bool(f_path is None or os.path.exists(f_path) is False):
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), f_path)
            else:
                return decorated(*args, **kwargs)

        return wrapper


def exception(logger):
    """
    A decorator that wraps the passed in function and logs
    exceptions should one occur

    param logger: The logging object
    type logger: logger
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except:
                # log the exception
                err = "There was an exception in  "
                err += func.__name__
                logger.exception(err, exc_info=True)
                # re-raise the exception
                raise

        return wrapper

    return decorator


class DebugDecorators(object):
    """File validation decorator

    Arguments:
        object {[type]} -- [description]

    Raises:
        None: 

    Returns:
        [] -- [Decorated function]
    """

    @classmethod
    def add_optional_debug_images_for_image_module(cls, decorated):
        """Add optional debug images in image_module class if DEBUG option is True in config
        """

        @functools.wraps(decorated)
        def wrapper(*args, **kwargs):
            """ wrapper for decorated function

            Arguments:
                cls {VideoDecorators} -- [Video decorators class]

            Returns:
                [function] -- [Decorated function with optional debug images if Debug option is on in config]
            """
            if config.Image.DEBUG == True:
                returnValue = decorated(*args, **kwargs)
                args[0].debug_images = args[0].crop_extractor.extracted_feature_maps
                if args[0].crop_selector.debug_image is not None:
                    args[0].debug_images.append(args[0].crop_selector.debug_image)
                return returnValue
            else:
                return decorated(*args, **kwargs)

        return wrapper



"""
.. module:: Katna.image_features.feature
    :platform: Cross Platform
    :synopsis: image_Feature.Feature is an Base class for Image feature\
         to be used for Katna Image crop functionality all image features\
              inherit from this base class
"""

import os
import cv2
import numpy as np


class Feature:
    """Base Class: Base class for all Image features: To be used \
        for Katna crop rectangle scoring, all image features inherit from \
            this base class
    """

    def __init__(self, weight):
        # Weight parameter for given feature
        self.weight = weight

    def get_weight(self):
        """gets weight of particular feature

        :return: weight of the feature
        :rtype: float
        """
        return self.weight



"""
.. module:: Katna.image
    :platform: OS X
    :synopsis: This module has functions related to smart cropping
"""
import os
import cv2
import numpy as np
from Katna.decorators import FileDecorators
from Katna.feature_list import FeatureList
from Katna.filter_list import FilterList
from Katna.crop_extractor import CropExtractor
from Katna.crop_selector import CropSelector
import Katna.config as config
from Katna.decorators import DebugDecorators


class UserFiltersEnum:
    """Enum class for filters"""

    text = "TextDetector"


class Image(object):
    """Class for all image cropping operations

    :param object: base class inheritance
    :type object: class:`Object`
    """

    def __init__(self, disable_text=True):
        """Constructor for image files"""
        featureList = FeatureList()
        filterList = FilterList()
        self.user_filters_enum = UserFiltersEnum()
        self.crop_extractor = CropExtractor()
        self.crop_selector = CropSelector()
        self.features = featureList.get_features()
        self.definedFilters = filterList.get_filters()

    def _get_crop_specs(
        self, image_height, image_width, ratio_height, ratio_width, is_height_small=True
    ):
        """Internal function to create the crop specs for a given aspect ratio

        :param image_height: height of image
        :type image_height: int, required
        :param image_width: width of image
        :type image_width: int, required
        :param ratio_height: aspect ratio height (eg. 3 from 4:3)
        :type ratio_height: int, required
        :param ratio_width: aspect ratio width (eg. 4 from 4:3)
        :type ratio_width: int, required
        :param is_height_small: parameter to check if crop dimension should be reduced wrt height[default=True]
        :type is_height_small: boolean, required
        :return: list of crop height and crop width
        :rtype:list of tuples
        """

        # multiplication factor by which height/width of crop should be decreased to get crop specs
        multiply_by = 1
        crop_list_tuple = []

        # Calculating the height and width ratio wrt aspect ratio
        hr, wr = image_height / ratio_height, image_width / ratio_width

        # print("hr, wr",hr, wr)
        # Check if height is smaller than the width.If yes, interchange height and width.
        if not is_height_small:
            image_height, image_width = image_width, image_height
            hr, wr = wr, hr
        crop_height, crop_width = image_height, hr * ratio_width

        # Decreasing the height and width for crops while checking it don't get small by 1/(min) of image height/width
        while True:
            if not (
                (crop_height >= (image_height // config.Image.min_image_to_crop_factor))
                and (
                    crop_width >= (image_width // config.Image.min_image_to_crop_factor)
                )
            ):
                break

            crop_height, crop_width = (
                int(crop_height),
                int((ratio_width / ratio_height) * crop_height),
            )

            crop_list_tuple.append((crop_height, crop_width))

            crop_height /= multiply_by

            crop_height, crop_width = (
                int(crop_height),
                int((ratio_width / ratio_height) * crop_height),
            )

            multiply_by += config.Image.crop_height_reduction_factor_in_each_iteration

        return crop_list_tuple

    # Apply optional Debug mode decorator , If config=DEBUG is true this decorator
    # will populate internal variables of Image module.debug_images with debug images
    # Which you can see by opencv Imshow to check if every feature is working as expected
    @DebugDecorators.add_optional_debug_images_for_image_module
    def crop_image_from_cvimage(
        self,
        input_image,
        crop_width,
        crop_height,
        num_of_crops,
        filters=[],
        down_sample_factor=config.Image.down_sample_factor,
    ):
        """smartly crops the imaged based on the specification - width and height

        :param input_image: Input image
        :type input_image: numpy array, required
        :param crop_width: output crop width
        :type crop_width: int
        :param crop_height: output crop heigh
        :type crop_height: int
        :param num_of_crops: number of crops required
        :type num_of_crops: int
        :param filters: filters to be applied for cropping(only returns crops containing english text where the crop rectangle doesn't cut the text)
        :type filters: list (eg. ['text'])
        :param down_sample_factor: number by which you want to reduce image height & width (use it if image is large or to fasten the process)
        :type down_sample_factor: int [default=8]
        :return: crop list
        :rtype: list of structure crop_rect
        """

        self.crop_extractor.down_sample_factor = down_sample_factor
        if (
            input_image.shape[0] + 5 <= crop_height
            or input_image.shape[1] + 5 <= crop_width
        ):
            # print(
            #     "Error: crop width or crop height larger than Image",
            #     "input_image.shape",
            #     input_image.shape,
            #     "crop_width",
            #     crop_width,
            #     "crop_height",
            #     crop_height,
            # )

            return []

        extracted_candidate_crops = self.crop_extractor.extract_candidate_crops(
            input_image, crop_width, crop_height, self.features
        )

        # print(extracted_candidate_crops)
        # text: TextDetector
        # dummy: DummyDetector
        self.filters = []
        for x in filters:
            try:
                self.filters.append(eval("self.user_filters_enum." + x))
            except AttributeError as e:
                print(str(e))
        # self.filters = [eval("user_filters_enum."+x) for x in filters]

        crops_list = self.crop_selector.select_candidate_crops(
            input_image,
            num_of_crops,
            extracted_candidate_crops,
            self.definedFilters,
            self.filters,
        )

        return crops_list

    @FileDecorators.validate_dir_path
    def crop_image_from_dir(
        self,
        dir_path,
        crop_width,
        crop_height,
        num_of_crops,
        filters=[],
        down_sample_factor=config.Image.down_sample_factor,
    ):
        """smartly crops all the images (inside a directory) based on the specification - width and height

        :param dir_path: Input Directory path
        :type dir_path: str, required
        :param crop_width: output crop width
        :type crop_width: int
        :param crop_height: output crop height
        :type crop_height: int
        :param num_of_crops: number of crops required
        :type num_of_crops: int
        :param filters: filters to be applied for cropping(checks if image contains english text and the crop rectangle doesn't cut the text)
        :type filters: list (eg. ['text'])
        :param down_sample_factor: number by which you want to reduce image height & width (use it if image is large or to fasten the process)
        :type down_sample_factor: int [default=8]
        :return: crop dict with key as filepath and crop list for the file
        :rtype: dict
        """

        all_crops = {}
        for path, subdirs, files in os.walk(dir_path):
            for filename in files:
                filepath = os.path.join(path, filename)
                if self._check_if_valid_image(filename):
                    image_file_path = os.path.join(path, filename)
                    crop_list = self.crop_image(
                        image_file_path,
                        crop_width,
                        crop_height,
                        num_of_crops,
                        filters,
                        down_sample_factor,
                    )
                    all_crops[filepath] = crop_list

        return all_crops

    @FileDecorators.validate_file_path
    def crop_image(
        self,
        file_path,
        crop_width,
        crop_height,
        num_of_crops,
        filters=[],
        down_sample_factor=config.Image.down_sample_factor,
    ):
        """smartly crops the imaged based on the specification - width and height

        :param file_path: Input file path
        :type file_path: str, required
        :param crop_width: output crop width
        :type crop_width: int
        :param crop_height: output crop heigh
        :type crop_height: int
        :param num_of_crops: number of crops required
        :type num_of_crops: int
        :param filters: filters to be applied for cropping(checks if image contains english text and the crop rectangle doesn't cut the text)
        :type filters: list (eg. ['text'])
        :param down_sample_factor: number by which you want to reduce image height & width (use it if image is large or to fasten the process)
        :type down_sample_factor: int [default=8]
        :return: crop list
        :rtype: list of structure crop_rect
        """

        imgFile = cv2.imread(file_path)

        crop_list = self.crop_image_from_cvimage(
            input_image=imgFile,
            crop_width=crop_width,
            crop_height=crop_height,
            num_of_crops=num_of_crops,
            filters=filters,
            down_sample_factor=down_sample_factor,
        )
        return crop_list

    @FileDecorators.validate_file_path
    def crop_image_with_aspect(
        self,
        file_path,
        crop_aspect_ratio,
        num_of_crops,
        filters=[],
        down_sample_factor=8,
    ):
        """smartly crops the imaged based on the aspect ratio and returns number of specified crops for each crop spec found in the image with
        the specified aspect ratio

        :param file_path: Input file path
        :type file_path: str, required
        :param crop_aspect_ratio: output crop ratio
        :type crop_aspect_ratio: str (eg. '4:3')
        :param num_of_crops: number of crops required
        :type num_of_crops: int
        :param filters: filters to be applied for cropping(checks if image contains english text and the crop rectangle doesn't cut the text)
        :type filters: list (eg. ['text'])
        :param down_sample_factor: number by which you want to reduce image height & width (use it if image is large or to fasten the process)
        :type down_sample_factor: int [default=8]
        :return: crop list
        :rtype: list of structure crop_rect
        """

        imgFile = cv2.imread(file_path)
        image_height, image_width, _ = imgFile.shape
        ratio_width, ratio_height = map(int, crop_aspect_ratio.split(":"))
        crop_list = self._generate_crop_options_given_for_given_aspect_ratio(
            imgFile,
            image_width,
            image_height,
            ratio_width,
            ratio_height,
            num_of_crops=num_of_crops,
            filters=filters,
            down_sample_factor=down_sample_factor,
        )

        sorted_list = sorted(crop_list, key=lambda x: float(x.score), reverse=True)
        return sorted_list[:num_of_crops]

    @FileDecorators.validate_file_path
    def save_crop_to_disk(self, crop_rect, frame, file_path, file_name, file_ext):
        """saves an in-memory crop on drive.

        :param crop_rect: In-memory crop_rect.
        :type crop_rect: crop_rect, required
        :param frame: In-memory input image.
        :type frame: numpy.ndarray, required
        :param file_name: name of the image.
        :type file_name: str, required
        :param file_path: Folder location where files needs to be saved
        :type file_path: str, required
        :param file_ext: File extension indicating the file type for example - '.jpg'
        :type file_ext: str, required
        :return: None
        """
        cropped_img = crop_rect.get_image_crop(frame)
        file_full_path = os.path.join(file_path, file_name + file_ext)
        cv2.imwrite(file_full_path, cropped_img)

    @FileDecorators.validate_file_path
    def resize_image(
        self,
        file_path,
        target_width,
        target_height,
        down_sample_factor=config.Image.down_sample_factor,
    ):
        """smartly resizes the image based on the specification - width and height

        :param file_path: Input file path
        :type file_path: str, required
        :param crop_width: output image width
        :type crop_width: int
        :param crop_height: output image heigh
        :type crop_height: int
        :param down_sample_factor: number by which you want to reduce image height & width (use it if image is large or to fasten the process)
        :type down_sample_factor: int [default=8]
        :return: resized image
        :rtype: cv_image
        """
        if not self._check_if_valid_image(file_path):
            print("Error: Invalid Image, check image path: ", file_path)
            return
        imgFile = cv2.imread(file_path)
        input_image_height, input_image_width, _ = imgFile.shape

        target_image_aspect_ratio = target_width / target_height
        input_image_aspect_ratio = input_image_width / input_image_height

        if input_image_aspect_ratio == target_image_aspect_ratio:
            target_image = cv2.resize(imgFile, (target_width, target_height))
            return target_image
        else:
            crop_list = self._generate_crop_options_given_for_given_aspect_ratio(
                imgFile,
                input_image_width,
                input_image_height,
                target_width,
                target_height,
                num_of_crops=1,
                filters=[],
                down_sample_factor=down_sample_factor,
            )
            # From list of crop options sort and get best crop using crop score variables in each
            # crop option
            sorted_list = sorted(crop_list, key=lambda x: float(x.score), reverse=True)
            # Get top crop image
            resized_image = sorted_list[0].get_image_crop(imgFile)
            target_image = cv2.resize(resized_image, (target_width, target_height))
            return target_image

    def _generate_crop_options_given_for_given_aspect_ratio(
        self,
        imgFile,
        input_image_width,
        input_image_height,
        target_width,
        target_height,
        num_of_crops,
        filters,
        down_sample_factor,
    ):
        """ Internal function to which for given aspect ratio (target_width/target_height)
        Generates ,scores and returns list of image crops 

        :param imgFile: Input image
        :type imgFile: opencv image
        :param input_image_width: input image width
        :type input_image_width: int
        :param input_image_height: input image height
        :type input_image_height: int
        :param target_width: target aspect ratio width
        :type target_width: int
        :param target_height: target aspect ratio height
        :type target_height: int
        :param num_of_crops: number of crop needed in the end
        :type num_of_crops: int
        :param filters: filters
        :type filters: list of filters
        :param down_sample_factor: image down sample factor for optimizing processing time
        :type down_sample_factor: int
        :return: list of candidate crop rectangles as per input aspect ratio
        :rtype: list of CropRect
        """
        crop_list_tuple, crop_list = [], []
        # Calculate height ratio and width ratio of input and target image
        height_ratio, width_ratio = (
            input_image_height / target_height,
            input_image_width / target_width,
        )

        # Generate candidate crops, _get_crop_spec function changes it's behavior based
        # on whether height_ratio is greater or smaller than width ratio.
        if height_ratio <= width_ratio:
            crop_list_tuple += self._get_crop_specs(
                input_image_height,
                input_image_width,
                target_height,
                target_width,
                is_height_small=True,
            )
        else:  # elif width_ratio < height_ratio:
            crop_list_tuple += self._get_crop_specs(
                input_image_height,
                input_image_width,
                target_height,
                target_width,
                is_height_small=False,
            )

        # For each of crop_specifications generated by _get_crop_spec() function
        # generate actual crop as well as give score to each of these crop 
        for crop_height, crop_width in crop_list_tuple:
            crop_list += self.crop_image_from_cvimage(
                input_image=imgFile,
                crop_width=crop_width,
                crop_height=crop_height,
                num_of_crops=num_of_crops,
                filters=filters,
                down_sample_factor=down_sample_factor,
            )
        return crop_list

    @FileDecorators.validate_dir_path
    def resize_image_from_dir(
        self,
        dir_path,
        target_width,
        target_height,
        down_sample_factor=config.Image.down_sample_factor,
    ):
        """smartly resizes all the images (inside a directory) based on the specification - width and height

        :param dir_path: Input Directory path
        :type dir_path: str, required
        :param target_width: output width
        :type target_width: int
        :param target_height: output height
        :type target_height: int
        :param down_sample_factor: number by which you want to reduce image height & width (use it if image is large or to fasten the process)
        :type down_sample_factor: int [default=8]
        :return: dict with key as filepath and resized image as in opencv format as value
        :rtype: dict
        """

        all_resized_images = {}
        for path, subdirs, files in os.walk(dir_path):
            for filename in files:
                filepath = os.path.join(path, filename)
                image_file_path = os.path.join(path, filename)
                if self._check_if_valid_image(image_file_path):
                    resized_image = self.resize_image(
                        image_file_path, target_width, target_height, down_sample_factor
                    )
                    all_resized_images[filepath] = resized_image
                else:
                    print("Error: Not a valid image file:", image_file_path)

        return all_resized_images

    @FileDecorators.validate_file_path
    def save_image_to_disk(self, image, file_path, file_name, file_ext):
        """saves an in-memory image obtained from image resize on drive.

        :param image: In-memory input image.
        :type image: numpy.ndarray, required
        :param file_name: name of the image.
        :type file_name: str, required
        :param file_path: Folder location where files needs to be saved
        :type file_path: str, required
        :param file_ext: File extension indicating the file type for example - '.jpg'
        :type file_ext: str, required
        :return: None
        """
        file_full_path = os.path.join(file_path, file_name + file_ext)
        cv2.imwrite(file_full_path, image)

    @FileDecorators.validate_file_path
    def _check_if_valid_image(self, file_path):
        """Function to check if given image file is a valid image compatible with
        opencv

        :param file_path: image filename
        :type file_path: str
        :return: Return True if valid image file else False
        :rtype: bool
        """
        try:
            frame = cv2.imread(file_path)
            # Making sure video frame is not empty
            if frame is not None:
                return True
            else:
                return False
        except cv2.error as e:
            print("cv2.error:", e)
            return False
        except Exception as e:
            print("Exception:", e)
            return False



"""
.. module:: Katna.image_selecter
    :platform: OS X
    :synopsis: This module has functions related to key frame extraction 
"""

from __future__ import print_function

import os
from glob import glob
import tempfile
import cv2
import numpy as np
from sklearn.cluster import KMeans
from skimage.filters.rank import entropy
from skimage.morphology import disk
from skimage import img_as_float
import itertools  
import time

import Katna.config as config


class ImageSelector(object):
    """Class for selection of best top N images from input list of images, Currently following selection method are implemented:
    brightness filtering, contrast/entropy filtering, clustering of frames and variance of laplacian for non blurred images 
    selection

    :param object: base class inheritance
    :type object: class:`Object`
    """

    def __init__(self, pool_obj):
        # Setting for Multiprocessing Pool Object
        self.pool_obj = pool_obj

        # Setting for optimum Brightness values
        self.min_brightness_value = config.ImageSelector.min_brightness_value
        self.max_brightness_value = config.ImageSelector.max_brightness_value
        self.brightness_step = config.ImageSelector.brightness_step 

        # Setting for optimum Contrast/Entropy values
        self.min_entropy_value = config.ImageSelector.min_entropy_value
        self.max_entropy_value = config.ImageSelector.max_entropy_value
        self.entropy_step = config.ImageSelector.entropy_step 

    def __get_brightness_score__(self, image):
        """Internal function to compute the brightness of input image , returns brightness score between 0 to 100.0 , 

        :param object: base class inheritance
        :type object: class:`Object`
        :param image: input image
        :type image: Opencv Numpy Image   
        :return: result of Brightness measurment 
        :rtype: float value between 0.0 to 100.0    
        """
        hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        _, _, v = cv2.split(hsv)
        sum = np.sum(v, dtype=np.float32)
        num_of_pixels = v.shape[0] * v.shape[1]
        brightness_score = (sum * 100.0) / (num_of_pixels * 255.0)
        return brightness_score

    def __get_entropy_score__(self, image):
        """Internal function to compute the entropy/contrast of input image , returns entropy score between 0 to 10 , 
 
        :param object: base class inheritance
        :type object: class:`Object`
        :param image: input image
        :type image: Opencv Numpy Image
        :return: result of Entropy measurment
        :rtype: float value between 0.0 to 10.0
        """
        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        entr_img = entropy(gray, disk(5))
        all_sum = np.sum(entr_img)
        num_of_pixels = entr_img.shape[0] * entr_img.shape[1]
        entropy_score = (all_sum) / (num_of_pixels)

        return entropy_score

    def __variance_of_laplacian__(self, image):
        """Internal function to compute the laplacian of the image and then return the focus
        measure, which is simply the variance of the laplacian,
 
        :param object: base class inheritance
        :type object: class:`Object`
        :param image: input image
        :type image: Opencv Numpy Image   
        :return: result of cv2.Laplacian
        :rtype: opencv image of type CV_64F    
        """

        return cv2.Laplacian(image, cv2.CV_64F).var()

    def __filter_optimum_brightness_and_contrast_images__(self, input_img_files):
        """ Internal function for selection of given input images with following parameters :optimum brightness and contrast range ,
        returns array of image files which are in optimum brigtness and contrast/entropy range.
 
        :param object: base class inheritance
        :type object: class:`Object`
        :param files: list of input image files 
        :type files: python list of images
        :return: Returns list of filtered images  
        :rtype: python list of images 
        """

        n_files = len(input_img_files)

        # -------- calculating the brightness and entropy score by multiprocessing ------
        brightness_score = np.array(
            self.pool_obj.map(self.__get_brightness_score__, input_img_files)
        )

        entropy_score = np.array(
            self.pool_obj.map(self.__get_entropy_score__, input_img_files)
        )

        # -------- Check if brightness and contrast scores are in the min and max defined range ------
        brightness_ok = np.where(
            np.logical_and(
                brightness_score > self.min_brightness_value,
                brightness_score < self.max_brightness_value,
            ),
            True,
            False,
        )
        contrast_ok = np.where(
            np.logical_and(
                entropy_score > self.min_entropy_value,
                entropy_score < self.max_entropy_value,
            ),
            True,
            False,
        )

        # Returning only thos images which are have good brightness and contrast

        return [
            input_img_files[i]
            for i in range(n_files)
            if brightness_ok[i] and contrast_ok[i]
        ]

    def __prepare_cluster_sets__(self, files):
        """ Internal function for clustering input image files, returns array of indexs of each input file
        (which determines which cluster a given file belongs)
 
        :param object: base class inheritance
        :type object: class:`Object`
        :param files: list of input image files 
        :type files: python list of opencv numpy images
        :return: Returns array containing index for each file for cluster belongingness 
        :rtype: np.array   
        """

        all_hists = []

        # Calculating the histograms for each image and adding them into **all_hists** list
        for img_file in files:
            img = cv2.cvtColor(img_file, cv2.COLOR_BGR2GRAY)
            hist = cv2.calcHist([img], [0], None, [256], [0, 256])
            hist = hist.reshape((256))
            all_hists.append(hist)

        # Kmeans clustering on the histograms
        kmeans = KMeans(n_clusters=self.nb_clusters, random_state=0).fit(all_hists)
        labels = kmeans.labels_

        # Identifying the label for each image in the cluster and tagging them
        files_clusters_index_array = []
        for i in np.arange(self.nb_clusters):
            index_array = np.where(labels == i)
            files_clusters_index_array.append(index_array)

        files_clusters_index_array = np.array(files_clusters_index_array)
        return files_clusters_index_array

    def __get_laplacian_scores(self, files, n_images):
        """Function to iteratee over each image in the cluster and calculates the laplacian/blurryness 
           score and adds the score to a list

        :param files: list of input filenames 
        :type files: python list of string
        :param n_images: number of images in the given cluster
        :type n_images: int
        :return: Returns list of laplacian scores for each image in the given cluster
        :rtype: python list 
        """

        variance_laplacians = []
        # Iterate over all images in image list
        for image_i in n_images:
            img_file = files[n_images[image_i]]
            img = cv2.cvtColor(img_file, cv2.COLOR_BGR2GRAY)

            # Calculating the blurryness of image
            variance_laplacian = self.__variance_of_laplacian__(img)
            variance_laplacians.append(variance_laplacian)

        return variance_laplacians

    def __get_best_images_index_from_each_cluster__(
        self, files, files_clusters_index_array
    ):
        """ Internal function returns index of one best image from each cluster

        :param object: base class inheritance
        :type object: class:`Object`
        :param files: list of input filenames 
        :type files: python list of string
        :param files_clusters_index_array: Input is array containing index for each file for cluster belongingness 
        :type: np.array   
        :return: Returns list of filtered files which are best candidate from each cluster
        :rtype: python list 
        """

        filtered_items = []

        # Iterating over every image in each cluster to find the best images from every cluster
        clusters = np.arange(len(files_clusters_index_array))
        for cluster_i in clusters:
            curr_row = files_clusters_index_array[cluster_i][0]
            # kp_lengths = []
            n_images = np.arange(len(curr_row))
            variance_laplacians = self.__get_laplacian_scores(files, n_images)

            # Selecting image with low burr(high laplacian) score
            selected_frame_of_current_cluster = curr_row[np.argmax(variance_laplacians)]
            filtered_items.append(selected_frame_of_current_cluster)

        return filtered_items

    def __getstate__(self):
        """Function to get the state of initialized class object and remove the pool object from it
        """
        self_dict = self.__dict__.copy()
        del self_dict["pool_obj"]
        return self_dict

    def __setstate__(self, state):
        """Function to update the state of initialized class object woth the pool object
        """
        self.__dict__.update(state)

    def select_best_frames(self, input_key_frames, number_of_frames):
        """[summary] Public function for Image selector class: takes list of key-frames images and number of required
        frames as input, returns list of filtered keyframes

        :param object: base class inheritance
        :type object: class:`Object`
        :param input_key_frames: list of input keyframes in list of opencv image format 
        :type input_key_frames: python list opencv images
        :param number_of_frames: Required number of images 
        :type: int   
        :return: Returns list of filtered image files 
        :rtype: python list of images
        """

        self.nb_clusters = number_of_frames

        filtered_key_frames = []
        filtered_images_list = []
        # Repeat until number of frames 
        min_brightness_values = np.arange(config.ImageSelector.min_brightness_value, -0.01, -self.brightness_step)
        max_brightness_values = np.arange(config.ImageSelector.max_brightness_value, 100.01, self.brightness_step)
        min_entropy_values = np.arange(config.ImageSelector.min_entropy_value, -0.01, -self.entropy_step)
        max_entropy_values = np.arange(config.ImageSelector.max_entropy_value, 10.01, self.entropy_step)

        for (min_brightness_value, max_brightness_value, min_entropy_value, max_entropy_value) in itertools.zip_longest(min_brightness_values, max_brightness_values, min_entropy_values, max_entropy_values): 
            if min_brightness_value is None:
                min_brightness_value = 0.0
            if max_brightness_value is None:
                max_brightness_value = 100.0
            if min_entropy_value is None:
                min_entropy_value = 0.0
            if max_entropy_value is None:
                max_entropy_value = 10.0
            self.min_brightness_value = min_brightness_value
            self.max_brightness_value = max_brightness_value
            self.min_entropy_value = min_entropy_value
            self.max_entropy_value = max_entropy_value
            filtered_key_frames = self.__filter_optimum_brightness_and_contrast_images__(
                input_key_frames, 
            )
            if len(filtered_key_frames) >= number_of_frames:
                break

        # Selecting the best images from each cluster by first preparing the clusters on basis of histograms 
        # and then selecting the best images from every cluster
        if len(filtered_key_frames) >= self.nb_clusters:
            files_clusters_index_array = self.__prepare_cluster_sets__(filtered_key_frames)
            selected_images_index = self.__get_best_images_index_from_each_cluster__(
                filtered_key_frames, files_clusters_index_array
            )

            for index in selected_images_index:
                img = filtered_key_frames[index]
                filtered_images_list.append(img)
        else:
            # if number of required files are less than requested key-frames return all the files
            for img in filtered_key_frames:
                filtered_images_list.append(img)
        return filtered_images_list



"""
.. module:: Katna.video_compressor
    :platform: OS X
    :synopsis: This module has functions related to video compression 
"""
import os
import cv2
import operator
import numpy as np
from scipy.signal import argrelextrema

import tempfile
import Katna.config as config

from Katna.decorators import VideoDecorators
from Katna.decorators import FileDecorators
import Katna.helper_functions as helper
import ffmpy
import ntpath


class VideoCompressor(object):
    """Class for performing video compression task: based on ffmpeg library for doing video compression"""

    def __init__(self):
        # Find out location of ffmpeg binary on system
        helper._set_ffmpeg_binary_path()

    @FileDecorators.validate_file_path
    def compress_video(
        self,
        file_path,
        force_overwrite,
        crf_parameter,
        output_video_codec,
        out_dir_path,
        out_file_name,
    ):
        """Function to compress given input file


        :param file_path: Input file path
        :type file_path: str        
        :param force_overwrite: optional parameter if True then if there is \
        already a file in output file location function will overwrite it, defaults to False
        :type force_overwrite: bool, optional
        :param crf_parameter: Constant Rate Factor Parameter for controlling \
        amount of video compression to be applied, The range of the quantizer scale is 0-51:\
        where 0 is lossless, 23 is default, and 51 is worst possible.\
        It is recommend to keep this value between 20 to 30 \
        A lower value is a higher quality, you can change default value by changing \
        config.Video.video_compression_crf_parameter
        :type crf_parameter: int, optional
        :param output_video_codec: Type of video codec to choose, \
        Currently supported options are libx264 and libx265, libx264 is default option.\
        libx264 is more widely supported on different operating systems and platforms, \
        libx265 uses more advanced x265 codec and results in better compression and even less \
        output video sizes with same or better quality. Right now libx265 is not as widely compatible \
        on older versions of MacOS and Widows by default. If wider video compatibility is your goal \
        you should use libx264., you can change default value by changing \
        Katna.config.Video.video_compression_codec
        :type output_video_codec: str, optional
        :param out_dir_path: output folder path where you want output video to be saved, defaults to ""
        :type out_dir_path: str, optional
        :param out_file_name: output filename, if not mentioned it will be same as input filename, defaults to ""
        :type out_file_name: str, optional
        :raises Exception: raises FileNotFoundError Exception if input video file not found, also exception is raised in case output video file path already exist and force_overwrite is not set to True.
        :return: Status code Returns True if video compression was successfull else False
        :rtype: bool
        """

        if not helper._check_if_valid_video(file_path):
            raise Exception("Invalid or corrupted video: " + file_path)

        output_full_file_name = self._generate_target_full_file_name(
            file_path, out_dir_path, out_file_name
        )

        if os.path.exists(output_full_file_name) is True and force_overwrite is False:
            exceptionMsg = (
                "Target video name already exist: "
                + output_full_file_name
                + "\nPass on parameter force_overwrite=True in video.compress_video() function to force overwrite"
            )
            raise Exception(exceptionMsg)

        status = self._compress_video_using_ffmpeg(
            file_path, crf_parameter, output_video_codec, output_full_file_name
        )
        return status

    def _generate_target_full_file_name(self, file_path, out_dir_path, out_file_name):
        """Internal function to generate output video file name given input video file_path, out_dir_path and out_file_name

        :param file_path: Input Video file path
        :type file_path: str
        :param out_dir_path: output folder path where you want output video to be saved, defaults to ""
        :type out_dir_path: str
        :param out_file_name: output filename, if not mentioned it will be same as input filename, defaults to ""
        :type out_file_name: str
        :return: Generated output video file path
        :rtype: str
        """
        # If output directory is not mentioned save compressed video in input dir
        if out_dir_path == "":
            out_dir_path = ntpath.dirname(file_path)
        if out_file_name != "":
            targetname = os.path.join(out_dir_path, out_file_name)
        else:
            # if output filename not mentioned make filename same as input file
            name, _ = os.path.splitext(ntpath.basename(file_path))
            targetname = os.path.join(out_dir_path, name)

        # Use mp4 extension for target regardless of input extension
        ext = config.Video.compression_output_file_extension
        target_full_file_name = "{0}.{1}".format(targetname, ext)

        # if target_file_name is same as input file name,
        # save it with extension "_compressed"
        if target_full_file_name == file_path:
            target_full_file_name = "{0}_compressed.{1}".format(targetname, ext)

        return target_full_file_name

    def _compress_video_using_ffmpeg(
        self, filename, crf_parameter, output_video_codec, targetname
    ):
        """Internal function to compress a input file with ffmpeg using given codec
        :param file_path: Input file path
        :type file_path: str
        :param crf_parameter: Constant Rate Factor Parameter for controlling amount of video compression to be applied, The range of the quantizer scale is 0-51: where 0 is lossless, 23 is default, and 51 is worst possible. A lower value is a higher quality, defaults to config.Video.video_compression_crf_parameter
        :type crf_parameter: int
        :param output_video_codec: Type of video codec to choose, Current Options: libx264 and libx265,libx264 is default and is more compatible to wide range of Operating system, libx265 gives better compression at expense of compatibility, defaults to config.Video.video_compression_codec
        :type output_video_codec: str
        :param targetname: path where output compressed video file to be stored
        :type targetname: str
        :return: Status code Returns True if video compression was successfull else False
        :rtype: bool
        """
        ffmpeg_output_parameters = (
            "-vcodec " + output_video_codec + " -crf " + str(crf_parameter)
        )
        # Uses ffmpeg binary for video clipping using ffmpy wrapper
        FFMPEG_BINARY = os.getenv("FFMPEG_BINARY")
        ff = ffmpy.FFmpeg(
            executable=FFMPEG_BINARY,
            inputs={filename: "-y -hide_banner -nostats -loglevel panic"},
            outputs={targetname: ffmpeg_output_parameters},
        )
        try:
            ff.run()
        except ffmpy.FFRuntimeError:
            return False
        return True



"""
.. module:: Katna.ffmpeg_helper
    :platform: OS X
    :synopsis: This module has functions related to key frame extraction 
"""
import os
import cv2
import operator
import numpy as np
from scipy.signal import argrelextrema

from Katna.decorators import VideoDecorators
from Katna.decorators import FileDecorators

import tempfile
import Katna.config as config
import subprocess
import re
import ffmpy
from imageio_ffmpeg import get_ffmpeg_exe


def _convert_to_seconds(time):
    """Will convert any time into seconds.
    If the type of `time` is not valid,
    it's returned as is.
    Here are the accepted formats::
    >>> convert_to_seconds(15.4)   # seconds
    15.4
    >>> convert_to_seconds((1, 21.5))   # (min,sec)
    81.5
    >>> convert_to_seconds((1, 1, 2))   # (hr, min, sec)
    3662
    >>> convert_to_seconds('01:01:33.045')
    3693.045
    >>> convert_to_seconds('01:01:33,5')    # coma works too
    3693.5
    >>> convert_to_seconds('1:33,5')    # only minutes and secs
    99.5
    >>> convert_to_seconds('33.5')      # only secs
    33.5

    :param time: time_string
    :type time: string
    :return: time in seconds
    :rtype: float
    """

    factors = (1, 60, 3600)

    if isinstance(time, str):
        time = [float(part.replace(",", ".")) for part in time.split(":")]

    if not isinstance(time, (tuple, list)):
        return time

    return sum(mult * part for mult, part in zip(factors, reversed(time)))


@FileDecorators.validate_file_path
def _check_if_valid_video(file_path):
    """Function to check if given video file is a valid video compatible with
    ffmpeg/opencv

    :param file_path: video filename
    :type file_path: str
    :return: Return True if valid video file else False
    :rtype: bool
    """
    try:
        vid = cv2.VideoCapture(file_path)
        if vid.isOpened():
            # Making sure we can read at least two frames from video
            ret, frame = vid.read()
            ret, frame = vid.read()
            # Making sure video frame is not empty
            if frame is not None:
                return True
            else:
                return False
        else:
            return False
    except cv2.error as e:
        print("cv2.error:", e)
        return False
    except Exception as e:
        print("Exception:", e)
        return False


def _set_ffmpeg_binary_path():
    """Function for getting path to ffmpeg binary on your system to be
    used by ffmpy
    # Derived from ffmpeg detection code borrowed from moviepy
    # https://github.com/Zulko/moviepy/moviepy/config.py
    # The MIT License (MIT)
    # Copyright (c) 2015 Zulko
    #
    :raises IOError: [description]
    """
    FFMPEG_BINARY = os.getenv("FFMPEG_BINARY", "ffmpeg-imageio")

    if FFMPEG_BINARY == "ffmpeg-imageio":
        FFMPEG_BINARY = get_ffmpeg_exe()
    else:
        success, err = _try_cmd([FFMPEG_BINARY])
        if not success:
            raise IOError(
                f"{err} - The path specified for the ffmpeg binary might be wrong"
            )
    os.environ["FFMPEG_BINARY"] = FFMPEG_BINARY


def _try_cmd(cmd):
    """
    # Derived from ffmpeg detection code borrowed from moviepy
    # https://github.com/Zulko/moviepy/moviepy/config.py
    # The MIT License (MIT)
    # Copyright (c) 2015 Zulko
    #
    :param cmd: command to execute
    :type cmd: string
    :return: True/False with error
    :rtype: Bool, Error
    """
    try:
        popen_params = _cross_platform_popen_params(
            {
                "stdout": subprocess.PIPE,
                "stderr": subprocess.PIPE,
                "stdin": subprocess.DEVNULL,
            }
        )
        proc = subprocess.Popen(cmd, **popen_params)
        proc.communicate()
    except Exception as err:
        return False, err
    else:
        return True, None


def _cross_platform_popen_params(popen_params):
    """
    # Derived from ffmpeg detection code borrowed from moviepy
    # https://github.com/Zulko/moviepy/moviepy/config.py
    # The MIT License (MIT)
    # Copyright (c) 2015 Zulko
    #
    Wrap with this function a dictionary of ``subprocess.Popen`` kwargs and
    will be ready to work without unexpected behaviours in any platform.
    Currently, the implementation will add to them:
    - ``creationflags=0x08000000``: no extra unwanted window opens on Windows
    when the child process is created. Only added on Windows.

    :param popen_params: original popen_parameters
    :type popen_params: dict
    :return: modified popen_parameters
    :rtype: dict
    """
    if os.name == "nt":
        popen_params["creationflags"] = 0x08000000
    return popen_params



"""
.. module:: Katna.video
    :platform: OS X
    :synopsis: This module has functions related to key frame extraction 
"""
import os.path
import os
import sys
import numpy as np
import cv2
import errno
from Katna.decorators import VideoDecorators
from Katna.decorators import FileDecorators

from Katna.frame_extractor import FrameExtractor
from Katna.image_selector import ImageSelector
from Katna.video_compressor import VideoCompressor
import Katna.helper_functions as helper

import Katna.config as config
import subprocess
import re
import ffmpy
from imageio_ffmpeg import get_ffmpeg_exe
from multiprocessing import Pool, Process, cpu_count

class Video(object):
    """Class for all video frames operations

    :param object: base class inheritance
    :type object: class:`Object`
    """

    def __init__(self):
        # Find out location of ffmpeg binary on system
        helper._set_ffmpeg_binary_path()
        # If the duration of the clipped video is less than **min_video_duration**
        # then, the clip will be added with the previous clipped
        self.min_video_duration = config.Video.min_video_duration
        # Calculating optimum number of processes for multiprocessing
        self.n_processes = cpu_count() // 2 - 1
        if self.n_processes < 1:
            self.n_processes = None

        # Folder to save the videos after clipping
        self.temp_folder = os.path.abspath(os.path.join("clipped"))
        if not os.path.isdir(self.temp_folder):
            os.mkdir(self.temp_folder)

    def _remove_clips(self, video_clips):
        """Remove video clips from the temp directory given list of video clips

        :param video_clips: [description]
        :type video_clips: [type]
        """
        for clip in video_clips:
            try:
                os.remove(clip)
            except OSError:
                print("Error in removing clip: " + clip)
            # print(clip, " removed!")

    @FileDecorators.validate_dir_path
    def extract_keyframes_from_videos_dir(self, no_of_frames, dir_path):
        """Returns best key images/frames from the videos in the given directory.
        you need to mention number of keyframes as well as directory path
        containing videos. Function returns python dictionary with key as filepath
        each dictionary element contains list of python numpy image objects as
        keyframes.

        :param no_of_frames: Number of key frames to be extracted
        :type no_of_frames: int, required
        :param dir_path: Directory location with videos
        :type dir_path: str, required
        :return: Dictionary with key as filepath and numpy.2darray Image objects
        :rtype: dict
        """

        all_videos_top_frames = {}

        for path, subdirs, files in os.walk(dir_path):
            for filename in files:
                filepath = os.path.join(path, filename)

                if helper._check_if_valid_video(filename):
                    video_file_path = os.path.join(path, filename)
                    video_top_frames = self.extract_video_keyframes(
                        no_of_frames, video_file_path
                    )
                    all_videos_top_frames[filepath] = video_top_frames

        return all_videos_top_frames

    @FileDecorators.validate_file_path
    def extract_video_keyframes(self, no_of_frames, file_path):
        """Returns a list of best key images/frames from a single video.

        :param no_of_frames: Number of key frames to be extracted
        :type no_of_frames: int, required
        :param file_path: video file location
        :type file_path: str, required
        :return: List of numpy.2darray Image objects
        :rtype: list
        """
        # Creating the multiprocessing pool
        self.pool_extractor = Pool(processes=self.n_processes)
        self.pool_selector = Pool(processes=self.n_processes)
        # Split the input video into chunks. Each split(video) will be stored
        # in a temp
        if not helper._check_if_valid_video(file_path):
            raise Exception("Invalid or corrupted video: " + file_path)

        chunked_videos = self._split(file_path)

        frame_extractor = FrameExtractor()

        # Passing all the clipped videos for  the frame extraction using map function of the
        # multiprocessing pool
        with self.pool_extractor:
            extracted_candidate_frames = self.pool_extractor.map(
                frame_extractor.extract_candidate_frames, chunked_videos
            )
            # Converting the nested list of extracted frames into 1D list
            extracted_candidate_frames = [
                frame for frames in extracted_candidate_frames for frame in frames
            ]

        self._remove_clips(chunked_videos)
        image_selector = ImageSelector(self.pool_selector)
        top_frames = image_selector.select_best_frames(
            extracted_candidate_frames, no_of_frames
        )

        return top_frames

    @FileDecorators.validate_file_path
    def compress_video(
        self,
        file_path,
        force_overwrite=False,
        crf_parameter=config.Video.video_compression_crf_parameter,
        output_video_codec=config.Video.video_compression_codec,
        out_dir_path="",
        out_file_name="",
    ):
        """Function to compress given input file

        :param file_path: Input file path
        :type file_path: str        
        :param force_overwrite: optional parameter if True then if there is \
        already a file in output file location function will overwrite it, defaults to False
        :type force_overwrite: bool, optional
        :param crf_parameter: Constant Rate Factor Parameter for controlling \
        amount of video compression to be applied, The range of the quantizer scale is 0-51:\
        where 0 is lossless, 23 is default, and 51 is worst possible.\
        It is recommend to keep this value between 20 to 30 \
        A lower value is a higher quality, you can change default value by changing \
        config.Video.video_compression_crf_parameter
        :type crf_parameter: int, optional
        :param output_video_codec: Type of video codec to choose, \
        Currently supported options are libx264 and libx265, libx264 is default option.\
        libx264 is more widely supported on different operating systems and platforms, \
        libx265 uses more advanced x265 codec and results in better compression and even less \
        output video sizes with same or better quality. Right now libx265 is not as widely compatible \
        on older versions of MacOS and Widows by default. If wider video compatibility is your goal \
        you should use libx264., you can change default value by changing \
        Katna.config.Video.video_compression_codec
        :type output_video_codec: str, optional
        :param out_dir_path: output folder path where you want output video to be saved, defaults to ""
        :type out_dir_path: str, optional
        :param out_file_name: output filename, if not mentioned it will be same as input filename, defaults to ""
        :type out_file_name: str, optional
        :raises Exception: raises FileNotFoundError Exception if input video file not found, also exception is raised in case output video file path already exist and force_overwrite is not set to True.
        :return: Status code Returns True if video compression was successfull else False
        :rtype: bool
        """
        # TODO add docstring for exeception 
        # Add details where libx265 will make sense
        
        if not helper._check_if_valid_video(file_path):
            raise Exception("Invalid or corrupted video: " + file_path)
        # Intialize video compression class
        video_compressor = VideoCompressor()
        # Run video compression
        status = video_compressor.compress_video(
            file_path,
            force_overwrite,
            crf_parameter,
            output_video_codec,
            out_dir_path,
            out_file_name,
        )
        return status

    @FileDecorators.validate_dir_path
    def compress_videos_from_dir(
        self,
        dir_path,
        force_overwrite=False,
        crf_parameter=config.Video.video_compression_crf_parameter,
        output_video_codec=config.Video.video_compression_codec,
        out_dir_path="",
        out_file_name="",
    ):
        """Function to compress input video files in a folder

        :param dir_path: Input folder path
        :type dir_path: str
        :param force_overwrite: optional parameter if True then if there is \
        already a file in output file location function will overwrite it, defaults to False
        :type force_overwrite: bool, optional
        :param crf_parameter: Constant Rate Factor Parameter for controlling \
        amount of video compression to be applied, The range of the quantizer scale is 0-51:\
        where 0 is lossless, 23 is default, and 51 is worst possible.\
        It is recommend to keep this value between 20 to 30 \
        A lower value is a higher quality, you can change default value by changing \
        config.Video.video_compression_crf_parameter
        :type crf_parameter: int, optional
        :param output_video_codec: Type of video codec to choose, \
        Currently supported options are libx264 and libx265, libx264 is default option.\
        libx264 is more widely supported on different operating systems and platforms, \
        libx265 uses more advanced x265 codec and results in better compression and even less \
        output video sizes with same or better quality. Right now libx265 is not as widely compatible \
        on older versions of MacOS and Widows by default. If wider video compatibility is your goal \
        you should use libx264., you can change default value by changing Katna.config.Video.video_compression_codec
        :type output_video_codec: str, optional
        :param out_dir_path: output folder path where you want output video to be saved, defaults to ""
        :type out_dir_path: str, optional
        :raises Exception: raises FileNotFoundError Exception if input video file not found, also exception is raised in case output video file path already exist and force_overwrite is not set to True.
        :return: Status code Returns True if video compression was successfull else False
        :rtype: bool
        """
        status = True
        list_of_videos_to_process = []
        # Collect all the valid video files inside folder
        for path, _, files in os.walk(dir_path):
            for filename in files:
                video_file_path = os.path.join(path, filename)
                if helper._check_if_valid_video(video_file_path):
                    list_of_videos_to_process.append(video_file_path)

        # Need to run in two sepearte loops to prevent recursion            
        for video_file_path in list_of_videos_to_process:
            statusI = self.compress_video(
                video_file_path,
                force_overwrite=force_overwrite,
                crf_parameter=crf_parameter,
                output_video_codec=output_video_codec,
                out_dir_path=out_dir_path,
            )
            status = bool(status and statusI)
        return status


    @FileDecorators.validate_file_path
    def save_frame_to_disk(self, frame, file_path, file_name, file_ext):
        """saves an in-memory numpy image array on drive.

        :param frame: In-memory image. This would have been generated by extract_video_keyframes method
        :type frame: numpy.ndarray, required
        :param file_name: name of the image.
        :type file_name: str, required
        :param file_path: Folder location where files needs to be saved
        :type file_path: str, required
        :param file_ext: File extension indicating the file type for example - '.jpg'
        :type file_ext: str, required
        :return: None
        """

        file_full_path = os.path.join(file_path, file_name + file_ext)
        cv2.imwrite(file_full_path, frame)

    @FileDecorators.validate_file_path
    def _split(self, file_path):
        """Function to split the videos and persist the chunks

        :param file_path: path of video file
        :type file_path: str, required
        :return: List of path of splitted video clips
        :rtype: list
        """
        clipped_files = []
        duration = self._get_video_duration_with_ffmpeg(file_path)
        # setting the start point to zero
        # Setting the breaking point for the clip to be 25 or if video is big
        # then relative to core available in the machine
        clip_start, break_point = (
            0,
            duration // cpu_count() if duration // cpu_count() > 15 else 25,
        )

        # Loop over the video duration to get the clip stating point and end point to split the video
        while clip_start < duration:

            clip_end = clip_start + break_point

            # Setting the end position of the particular clip equals to the end time of original clip,
            # if end position or end position added with the **min_video_duration** is greater than
            # the end time of original video
            if clip_end > duration or (clip_end + self.min_video_duration) > duration:
                clip_end = duration

            clipped_files.append(self._write_videofile(file_path, clip_start, clip_end))

            clip_start = clip_end
        return clipped_files

    def _write_videofile(self, video_file_path, start, end):
        """Function to clip the video for given start and end points and save the video

        :param video_file_path: path of video file
        :type video_file_path: str, required
        :param start: start time for clipping
        :type start: float, required
        :param end: end time for clipping
        :type end: float, required
        :return: path of splitted video clip
        :rtype: str
        """

        name = os.path.split(video_file_path)[1]

        # creating a unique name for the clip video
        # Naming Format: <video name>_<start position>_<end position>.mp4
        _clipped_file_path = os.path.join(
            self.temp_folder,
            "{0}_{1}_{2}.mp4".format(
                name.split(".")[0], int(1000 * start), int(1000 * end)
            ),
        )

        self._ffmpeg_extract_subclip(
            video_file_path, start, end, targetname=_clipped_file_path
        )
        return _clipped_file_path

    def _ffmpeg_extract_subclip(self, filename, t1, t2, targetname=None):
        """chops a new video clip from video file ``filename`` between
            the times ``t1`` and ``t2``, Uses ffmpy wrapper on top of ffmpeg
            library
        :param filename: path of video file
        :type filename: str, required
        :param t1: time from where video to clip
        :type t1: float, required
        :param t2: time to which video to clip
        :type t2: float, required
        :param targetname: path where clipped file to be stored
        :type targetname: str, optional
        :return: None
        """
        name, ext = os.path.splitext(filename)

        if not targetname:
            T1, T2 = [int(1000 * t) for t in [t1, t2]]
            targetname = name + "{0}SUB{1}_{2}.{3}".format(name, T1, T2, ext)

        timeParamter = "-ss " + "%0.2f" % t1 + " -t " + "%0.2f" % (t2 - t1)

        # Uses ffmpeg binary for video clipping using ffmpy wrapper
        FFMPEG_BINARY = os.getenv("FFMPEG_BINARY")
        ff = ffmpy.FFmpeg(
            executable=FFMPEG_BINARY,
            inputs={filename: "-y -hide_banner -loglevel panic "},
            outputs={targetname: timeParamter + " -vcodec copy -acodec copy"},
        )
        ff.run()

    @FileDecorators.validate_file_path
    def _get_video_duration_with_ffmpeg(self, file_path):
        """Get video duration using ffmpeg binary.
        Based on function ffmpeg_parse_infos inside repo
        https://github.com/Zulko/moviepy/moviepy/video/io/ffmpeg_reader.py
        The MIT License (MIT)
        Copyright (c) 2015 Zulko
        Returns a video duration in second

        :param file_path: video file path
        :type file_path: string
        :raises IOError:
        :return: duration of video clip in seconds
        :rtype: float
        """
        FFMPEG_BINARY = os.getenv("FFMPEG_BINARY")
        # Open the file in a pipe, read output
        ff = ffmpy.FFmpeg(
            executable=FFMPEG_BINARY,
            inputs={file_path: ""},
            outputs={None: "-f null -"},
        )
        _, error = ff.run(
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        infos = error.decode("utf8", errors="ignore")
        lines = infos.splitlines()
        if "No such file or directory" in lines[-1]:
            raise IOError(
                (
                    "Error: the file %s could not be found!\n"
                    "Please check that you entered the correct "
                    "path."
                )
                % file_path
            )

        # get duration (in seconds) by parsing ffmpeg file info returned by
        # ffmpeg binary
        video_duration = None
        decode_file = False
        try:
            if decode_file:
                line = [line for line in lines if "time=" in line][-1]
            else:
                line = [line for line in lines if "Duration:" in line][-1]
            match = re.findall("([0-9][0-9]:[0-9][0-9]:[0-9][0-9].[0-9][0-9])", line)[0]
            video_duration = helper._convert_to_seconds(match)
        except Exception:
            raise IOError(
                f"error: failed to read the duration of file {file_path}.\n"
                f"Here are the file infos returned by ffmpeg:\n\n{infos}"
            )
        return video_duration
